# -*- coding: utf-8 -*-
"""
oathldap_web.views - method view classes
"""

import socket
import smtplib
import hashlib
import time
import email.utils

import ldap0
import ldap0.filter
import ldap0.err
from ldap0 import LDAPError
from ldap0.ldapobject import ReconnectLDAPObject
from ldap0.controls.sessiontrack import SessionTrackingControl, SESSION_TRACKING_FORMAT_OID_USERNAME
from ldap0.pw import random_string

import mailutil

from flask import current_app, request, render_template
from flask.views import MethodView

from . import RequestLogAdaptor, ExtLDAPUrl
from .forms import TokenResetForm


class Default(MethodView):
    """
    Handle requests to base URL
    """

    def __init__(self, *args, **kwargs):
        MethodView.__init__(self, *args, **kwargs)
        self.logger = RequestLogAdaptor(
            current_app.logger,
            {
                'remote_ip': request.remote_addr,
                'req_class': '.'.join((self.__class__.__module__, self.__class__.__name__)),
                'req_id': id(request),
            }
        )
        self.logger.debug(
            '%s request from %s (via %s)',
            request.method,
            request.remote_addr,
            '>'.join(request.access_route),
        )
        self.ldap_conn = None
        self.user_ldap_conn = None

    def get(self):
        """
        Simply display the entry landing page
        """
        return render_template('default.html')


class BaseApp(Default):
    """
    Request handler base class which is not used directly
    """

    def _sess_track_ctrl(self):
        """
        return LDAPv3 session tracking control representing current user
        """
        return SessionTrackingControl(
            request.remote_addr,
            request.host,
            SESSION_TRACKING_FORMAT_OID_USERNAME,
            str(id(self)),
        )

    def ldap_connect(self, authz_id=None):
        """
        Connect and bind to the LDAP directory as local system account
        """
        self.ldap_url = ExtLDAPUrl(current_app.config['LDAP_URL'])
        self.ldap_conn = ReconnectLDAPObject(
            self.ldap_url.connect_uri(),
            trace_level=current_app.config['LDAP0_TRACE_LEVEL'],
        )
        # Send SASL bind request with mechanism EXTERNAL
        self.ldap_conn.sasl_non_interactive_bind_s('EXTERNAL', authz_id=authz_id)
        # end of ldap_connect()

    def open_user_conn(self, username, password):
        """
        Search a user entry specified by :username: and check
        :password: with LDAP simple bind.
        """
        self.user_ldap_conn = None
        user = self.ldap_conn.find_unique_entry(
            self.ldap_url.dn,
            scope=self.ldap_url.scope,
            filterstr=current_app.config['FILTERSTR_ADMIN_LOGIN'].format(uid=username),
            attrlist=['1.1'],
        )
        self.user_ldap_conn = ReconnectLDAPObject(
            self.ldap_url.connect_uri(),
            trace_level=current_app.config['LDAP0_TRACE_LEVEL'],
        )
        self.user_ldap_conn.simple_bind_s(user.dn_s, password.encode('utf-8'))
        # end of BaseApp.open_user_conn()

    def search_token(self, token_serial):
        """
        Search a token entry specified by serial number
        """
        token = self.user_ldap_conn.find_unique_entry(
            self.ldap_url.dn,
            scope=self.ldap_url.scope,
            filterstr=current_app.config['FILTERSTR_TOKEN_SEARCH'].format(
                owner_attr=current_app.config['ATTR_OWNER_DN'],
                serial=token_serial,
            ),
            attrlist=[
                'createTimestamp',
                'displayName',
                'oathFailureCount',
                'oathHOTPCounter',
                'oathHOTPParams',
                'oathLastFailure',
                'oathLastLogin',
                'oathSecretTime',
                'oathTokenIdentifier',
                'oathTokenSerialNumber',
                current_app.config['ATTR_OWNER_DN'],
            ],
            req_ctrls=[self._sess_track_ctrl()],
        )
        return token.entry_s['displayName'][0], token.dn_s, token.entry_s
        # endof BaseApp.search_token()

    def do_the_work(self):
        """
        this method contains the real work and is implemented by derived classes
        """
        raise NotImplementedError(
            "method .do_the_work() not implemented in class %s.%s" % (
                self.__class__.__module__,
                self.__class__.__name__,
            )
        )


    def clean_up(self):
        """
        Clean up initialized stuff
        """
        for conn in (self.ldap_conn, self.user_ldap_conn):
            if conn:
                try:
                    self.ldap_conn.unbind_s()
                except (AttributeError, LDAPError):
                    pass
        # end of BaseApp.clean_up()

    def post(self):
        """
        Process a POST request likely resulting in some write access

        In this wrapper method only form is validated and LDAP connection
        is opened. Afterwards self.do_the_work() is called which does the
        real use-case specific work.
        """
        # Parse and validate the form input
        self.form = TokenResetForm(request.form, csrf_enabled=False)
        if not self.form.validate():
            self.logger.error(
                'Input form data not valid (%s): %s',
                self.form.__class__.__name__,
                self.form.errors,
            )
            return render_template('reset_form.html', message='Incomplete or invalid input!')
        # Make connection to LDAP server
        try:
            self.ldap_connect(authz_id=current_app.config['LDAPI_AUTHZ_ID'])
        except LDAPError as ldap_err:
            self.logger.error(
                'LDAPError connecting to %r: %s',
                self.ldap_url.connect_uri(),
                ldap_err,
            )
            return render_template('default.html', message='Internal LDAP error!')
        # Do the real work
        try:
            res = self.do_the_work()
        except Exception as err:
            self.logger.error('Unhandled exception: %s', err, exc_info=__debug__)
            res = render_template('default.html', message='Internal error!')
        self.clean_up()
        return res
        # end of BaseApp.POST()


class ResetToken(BaseApp):
    """
    Resets token to unusable state but with temporary enrollment password.

    LDAP operations are authenticated with LDAPI/SASL/EXTERNAL
    """

    def get(self):
        """
        Process the GET request mainly for displaying input form
        """
        if 'serial' not in request.args:
            message = 'Enter a serial number of token to be (re-)initialized.'
        elif 'admin' not in request.args:
            message = 'Login with your 2FA admin account.'
        return render_template(
            'reset_form.html',
            message=message,
            admin=request.args.get('admin', ''),
            serial=request.args.get('serial', ''),
        )

    def _send_pw(self, token_serial, owner_entry, enroll_pw1):
        """
        Send 2nd part of temporary password to token owner
        """

        # Open connection to SMTP relay
        #---------------------------------------------------------------
        smtp_conn = mailutil.smtp_connection(
            current_app.config['SMTP_URL'],
            local_hostname=current_app.config['SMTP_LOCALHOSTNAME'],
            ca_certs=current_app.config['SMTP_TLS_CACERTS'],
            debug_level=current_app.config['SMTP_DEBUGLEVEL'],
        )

        # Construct the message
        #---------------------------------------------------------------
        smtp_message_tmpl = open(current_app.config['EMAIL_TEMPLATE'], 'rb').read().decode('utf-8')
        to_addr = owner_entry['mail'][0]
        default_headers = (
            ('From', current_app.config['SMTP_FROM']),
            ('Date', email.utils.formatdate(time.time(), True)),
        )
        owner_data = {
            'serial': token_serial,
            'admin': self.form.admin.data,
            'enrollpw1': enroll_pw1,
            'remote_ip': request.remote_addr,
            'fromaddr': current_app.config['SMTP_FROM'],
        }
        smtp_message = smtp_message_tmpl % owner_data
        smtp_subject = current_app.config['EMAIL_SUBJECT'] % owner_data

        # Send the message
        #---------------------------------------------------------------
        smtp_conn.send_simple_message(
            current_app.config['SMTP_FROM'],
            [to_addr],
            'utf-8',
            default_headers+(
                ('Subject', smtp_subject),
                ('To', to_addr),
            ),
            smtp_message,
        )
        smtp_conn.quit()
        self.logger.info('Sent reset password to %r.', to_addr)
        # end of _send_pw()

    def search_accounts(self, token_dn):
        """
        Search all accounts using the token
        """
        ldap_result = self.user_ldap_conn.search_s(
            self.ldap_url.dn,
            ldap0.SCOPE_SUBTREE,
            filterstr='(&(objectClass=oathUser)(oathToken={dn}))'.format(
                dn=ldap0.filter.escape_str(token_dn),
            ),
            attrlist=['uid', 'description']
        )
        if not ldap_result:
            return None
        return [
            (
                res.entry_s['uid'][0],
                res.entry_s.get('description', [''])[0],
            )
            for res in ldap_result
        ]

    def read_owner(self, owner_dn):
        """
        Read a token owner entry
        """
        ldap_result = self.user_ldap_conn.read_s(
            owner_dn,
            filterstr=current_app.config['FILTERSTR_OWNER_READ'],
            attrlist=[
                'displayName',
                'mail',
                'telePhoneNumber',
                'mobile',
                'l',
            ],
        )
        if ldap_result:
            result = ldap_result.entry_s
        else:
            raise ldap0.NO_SUCH_OBJECT(
                'No result with %r' % (current_app.config['FILTERSTR_OWNER_READ'],)
            )
        return result
        # end of read_owner()

    def update_token(self, token_dn, token_entry, token_password):
        """
        Resets token to unusable state by
        - overwriting 'oathSecret'
        - removing 'oathLastLogin'
        - removing 'oathHOTPCounter'
        - removing failure attributes 'oathFailureCount' and 'oathLastFailure'
        - setting temporary enrollment password in 'userPassword'
        - resetting 'oathSecretTime' to current time
        """
        session_tracking_ctrl = self._sess_track_ctrl()
        token_mods = [
            # We don't fully trust enrollment client
            # => set shared secret time to current time here
            (
                ldap0.MOD_REPLACE,
                b'oathSecretTime',
                [time.strftime('%Y%m%d%H%M%SZ', time.gmtime(time.time())).encode('ascii')],
            ),
        ]
        for del_attr in (
                'oathHOTPCounter',
                'oathLastLogin',
                'oathFailureCount',
                'oathLastFailure',
            ):
            if del_attr in token_entry:
                token_mods.append(
                    (ldap0.MOD_DELETE, del_attr.encode('ascii'), None)
                )
        # Reset the token entry
        self.user_ldap_conn.modify_s(
            token_dn,
            token_mods,
            req_ctrls=[session_tracking_ctrl],
        )
        # Try to remove shared secret separately because with
        # strict access control we don't know whether it's set or not
        try:
            self.user_ldap_conn.modify_s(
                token_dn,
                [(ldap0.MOD_DELETE, b'oathSecret', None)],
                req_ctrls=[session_tracking_ctrl],
            )
        except ldap0.NO_SUCH_ATTRIBUTE:
            # We can happily ignore this case
            pass
        # Set the new userPassword with Modify Password ext.op.
        # for server-side hashing
        self.ldap_conn.passwd_s(
            token_dn,
            None, token_password,
            req_ctrls=[session_tracking_ctrl],
        )
        # end of ResetToken.update_token()

    def do_the_work(self):
        """
        Actually do the work herein
        """
        # Check the user login and open user connection
        self.logger.debug('Will try login as user %r', self.form.admin.data)
        try:
            self.open_user_conn(self.form.admin.data, self.form.password.data)
        except LDAPError as ldap_err:
            self.logger.error(
                'Error opening user connection to %r as user %r: %s',
                self.ldap_url.connect_uri(),
                self.form.admin.data,
                ldap_err,
            )
            return render_template('reset_form.html', message='Admin login failed!')
        token_serial = self.form.serial.data
        try:
            token_displayname, token_dn, token_entry = self.search_token(
                token_serial
            )
            owner_dn = token_entry[current_app.config['ATTR_OWNER_DN']][0]
            owner_entry = self.read_owner(owner_dn)
            accounts = self.search_accounts(token_dn)
            confirm_hash = hashlib.sha256(
                ''.join((
                    token_serial,
                    owner_dn,
                    str(sorted(accounts or [])),
                )).encode('utf-8')
            ).hexdigest()
            if self.form.confirm.data != confirm_hash:
                return render_template(
                    'reset_form.html',
                    message='Please confirm token reset. Examine this information carefully!',
                    admin=self.form.admin.data,
                    serial=self.form.serial.data,
                    token=token_displayname,
                    owner=owner_entry['displayName'][0],
                    email=owner_entry['mail'][0],
                    accounts=accounts,
                    confirm=confirm_hash,
                )
            enroll_pw1 = random_string(
                alphabet=current_app.config['PWD_TMP_CHARS'],
                length=current_app.config['PWD_LENGTH']-current_app.config['PWD_ADMIN_LEN'],
                )
            enroll_pw2 = random_string(
                alphabet=current_app.config['PWD_TMP_CHARS'],
                length=current_app.config['PWD_ADMIN_LEN'],
            )
            enroll_pw = enroll_pw1 + enroll_pw2
            self.update_token(token_dn, token_entry, enroll_pw)
        except ldap0.err.NoUniqueEntry as ldap_err:
            self.logger.warning('LDAPError: %s', ldap_err)
            return render_template('reset_form.html', message='Serial no. not found!')
        except LDAPError as ldap_err:
            self.logger.error('LDAPError: %s', ldap_err, exc_info=__debug__)
            return render_template('reset_form.html', message='Internal LDAP error!')
        except Exception as err:
            self.logger.error('Unhandled exception: %s', err, exc_info=__debug__)
            return render_template('reset_form.html', message='Internal error!')
        # try to send 2nd enrollment password part to token owner
        try:
            self._send_pw(self.form.serial.data, owner_entry, enroll_pw1)
        except (socket.error, socket.gaierror, smtplib.SMTPException) as mail_error:
            self.logger.error(
                'Error sending e-mail to owner of device %s: %s',
                self.form.serial.data,
                mail_error,
                exc_info=__debug__
            )
            return render_template('reset_form.html', message='Error sending e-mail via SMTP!')
        self.logger.info('Finished resetting token %r.', token_dn)
        return render_template(
            'reset_action.html',
            message='Token was reset',
            serial=token_serial,
            token=token_entry['displayName'][0],
            owner=owner_entry['displayName'][0],
            email=owner_entry['mail'][0],
            enrollpw2=enroll_pw2,
        )
        # end of ResetToken.do_the_work()



# -*- coding: utf-8 -*-
"""
Configuration module for oathenroll
"""

import secrets

# URL path prefix used when generating URLs in e-mails
# handy for reverse proxy setups
APPLICATION_ROOT = '/oath'

# generate a 64-bytes random secret
APP_SECRET = secrets.token_bytes(64)

# default log level
LOG_LEVEL = 'info'

# logger name
LOG_NAME = 'oathldap_web'

# path of logging config file
LOG_CONFIG = None

# number of proxy levels
# see https://werkzeug.palletsprojects.com/en/1.0.x/middleware/proxy_fix/
PROXY_LEVEL = 0

# Trace level for ldap0 logs
LDAP0_TRACE_LEVEL = 0

# LDAP-URL describing the connection parameters and bind information
LDAP_URL = 'ldapi:///ou=ae-dir??sub??trace=0,x-saslmech=EXTERNAL'

# SASL authz-ID to be sent along with SASL/EXTERNAL bind
LDAPI_AUTHZ_ID = None

# Filter string template for finding an active admin entry during login
# Notes:
# - must require initialized 2FA user account with filter part (&(objectClass=oathUser)(oathHOTPToken=*))
# - authorization is enforced via filter part (|(memberOf=cn=otp-zone-admins,cn=otp,ou=ae-dir)(memberOf=cn=ae-admins,cn=ae,ou=ae-dir))
FILTERSTR_ADMIN_LOGIN = (
    '(&(objectClass=aeUser)(uid={uid})'
#    '(objectClass=oathUser)(oathHOTPToken=*)'  # enforces 2FA user account
    ')'
)

ATTR_OWNER_DN = 'owner'

# Filter string template for reading a particular active owner entry
# (require initialized 2FA user account herein)
FILTERSTR_OWNER_READ = '(&(objectClass=inetOrgPerson)(aeStatus=0))'

# Filter string template for finding an active token entry
FILTERSTR_TOKEN_SEARCH = '(&(objectClass=device)({owner_attr}=*)(objectClass=oathToken)(serialNumber=yubikey-{serial})(aeHwSerialNumber={serial})(oathTokenSerialNumber={serial}))'

# Name of directory containing all the template files
TEMPLATES_DIRNAME = '/etc/oath-ldap/oathenroll/templates/'

# Format string for displaying date and time
TIME_DISPLAY_FORMAT = '%Y-%m-%d %H:%M:%S'

# Length of generated temporary passwords
PWD_LENGTH = 12

# Number of chars of generated temporary passwords to be displayed to 2FA admin
PWD_ADMIN_LEN = 6

# Characters used for the temporary passwords
PWD_TMP_CHARS = 'abcdefghijkmnopqrstuvwxyzABCDEFGHIJKLMNPQRSTUVWXYZ23456789'

# Filename of template for sending e-mail message to user
EMAIL_SUBJECT = 'Your temporary enrollment password for Yubikey #%(serial)s'
EMAIL_TEMPLATE = TEMPLATES_DIRNAME+'reset.txt'

# SMTP server used as smart host (SMTP relay)
SMTP_URL = 'smtp://mail.example.com/?STARTTLS'

# Debug level for SMTP messages sent to stderr
SMTP_DEBUGLEVEL = 0

# Hostname to be sent in EHLO request,
# set to None for automatically using the local hostname
SMTP_LOCALHOSTNAME = 'ae-dir-suse-p1.vnet1.local'

# Path name of file containing CA certificates used to validate TLS server certs
SMTP_TLS_CACERTS = '/opt/ae-dir/etc/my-ae-dir-testca-2017-06.pem'

# From address in sent e-mail
SMTP_FROM = 'oath-ldap-admins@example.com'



# -*- coding: utf-8 -*-
"""
oathldap_web - OATH-LDAP token enrollment
"""

import os
import logging
import logging.config
from urllib.parse import quote_plus as url_quote_plus

from ldap0.ldapurl import LDAPUrl

#---------------------------------------------------------------------------
# constants
#---------------------------------------------------------------------------

# more complex HTTP security header values
HTTP_CSP_HEADER = ' '.join((
    "base-uri 'none';",
    "child-src 'none';",
    "connect-src 'none';",
    "default-src 'none';",
    "font-src 'self';",
    "form-action 'self';",
    "frame-ancestors 'none';",
    "frame-src 'none';",
    "img-src 'self' data:;",
    "media-src 'none';",
    "object-src 'none';",
    "script-src 'none';",
    "style-src 'self';",
    "require-trusted-types-for 'script';",
))
HTTP_PERMISSIONS_POLICY_HEADER = ', '.join((
    'accelerometer=(none)',
    'ambient-light-sensor=(none)',
    'camera=(none)',
    'clipboard-read=(none)',
    'clipboard-write=(none)',
    'display-capture'
    'geolocation=(none)',
    'gyroscope=(none)',
    'magnetometer=(none)',
    'microphone=(none)',
    'midi=(none)',
    'notifications=(none)',
    'push=(none)',
    'speaker-selection=(none)',
))

HTTP_HEADERS = (
    ('Cache-Control', 'no-store,no-cache,max-age=0,must-revalidate'),
    ('X-XSS-Protection', '1; mode=block'),
    ('X-DNS-Prefetch-Control', 'off'),
    ('X-Content-Type-Options', 'nosniff'),
    ('X-Frame-Options', 'deny'),
    ('Server', 'unknown'),
    ('Content-Security-Policy', HTTP_CSP_HEADER),
    ('X-Webkit-CSP', HTTP_CSP_HEADER),
    ('X-Content-Security-Policy', HTTP_CSP_HEADER),
    ('Referrer-Policy', 'same-origin'),
    ('Permissions-Policy', HTTP_PERMISSIONS_POLICY_HEADER),
    ('Cross-Origin-Embedder-Policy', 'require-corp'),
    ('Cross-Origin-Opener-Policy', 'same-origin'),
    ('Cross-Origin-Resource-Policy', 'same-site'),
)

#---------------------------------------------------------------------------
# basic functions and classes
#---------------------------------------------------------------------------

class ExtLDAPUrl(LDAPUrl):
    """
    Special class for handling additional LDAP URL extensions
    """
    attr2extype = {
        'who': 'bindname',
        'cred': 'X-BINDPW',
        'start_tls': 'startTLS',
        'trace_level': 'trace',
    }


class RequestLogAdaptor(logging.LoggerAdapter):
    """
    wrapper for adding more request-specific information to log messages
    """

    def process(self, msg, kwargs):
        return (
            'IP=%s CLASS=%s REQID=%d - %s' % (
                self.extra['remote_ip'],
                self.extra['req_class'],
                self.extra['req_id'],
                msg,
            ),
            kwargs,
        )


def validate_config(cfg):
    """
    validate some config parameters
    """
    # Safety check for URL chars
    if cfg['PWD_TMP_CHARS'] != url_quote_plus(cfg['PWD_TMP_CHARS']):
        raise ValueError(
            'URL special chars in PWD_TMP_CHARS: %r' % (cfg['PWD_TMP_CHARS'],)
        )



# -*- coding: utf-8 -*-
"""
package/install module package oathldap
"""

import sys
import os
from setuptools import setup, find_packages

PYPI_NAME = 'oath-ldap-web'

BASEDIR = os.path.dirname(os.path.realpath(__file__))

sys.path.insert(0, os.path.join(BASEDIR, 'oathldap_web'))
import __about__

setup(
    name=PYPI_NAME,
    license=__about__.__license__,
    version=__about__.__version__,
    description='OATH-LDAP web application',
    author=__about__.__author__,
    author_email=__about__.__mail__,
    maintainer=__about__.__author__,
    maintainer_email=__about__.__mail__,
    url='https://oath-ldap.stroeder.com',
    download_url='https://pypi.org/project/{0}/#files'.format(PYPI_NAME),
    keywords=['LDAP', 'OpenLDAP', 'OATH', 'OATH-LDAP', 'HOTP', 'TOTP', 'Yubikey'],
    packages=find_packages(exclude=['tests']),
    package_dir={'': '.'},
    #test_suite='tests',
    python_requires='>=3.6',
    include_package_data=True,
    data_files=[],
    install_requires=[
        'setuptools',
        'ldap0 >= 1.2.1',
        'mailutil',
        'Werkzeug>=1.0.0',
        'Flask',
        'WTForms',
        'Jinja2',
    ],
    zip_safe=False,
)



# -*- coding: utf-8 -*-
"""
oathldap_web.forms - forms declarations
"""

from wtforms import (
    Form,
    StringField,
    PasswordField,
    SubmitField,
    validators,
)


class TokenResetForm(Form):
    """
    token reset form
    """
    admin = StringField(
        'Admin username',
        [
            validators.Length(min=2, max=64),
            validators.InputRequired(),
            validators.Regexp('^[a-zA-Z0-9._-]+$'),
        ],
    )
    password = PasswordField(
        'New Password',
        [
            validators.InputRequired(),
        ],
    )
    serial = StringField(
        'Device serial',
        [
            validators.Length(min=4, max=32),
            validators.InputRequired(),
            validators.Regexp('^[0-9]*$'),
        ],
    )
    confirm = StringField(
        'Confirmation hash',
        [
            validators.Optional(strip_whitespace=True),
            validators.Regexp('^[0-9a-fA-F]*$'),
        ],
    )
    submit = SubmitField()



# -*- coding: utf-8 -*-
"""
oathldap_web.app - the application entry point
"""

import os
import logging
import logging.config

from werkzeug.middleware.proxy_fix import ProxyFix

import flask.logging
from flask import Flask

from . import validate_config, HTTP_HEADERS
from .views import Default, ResetToken


def create_app(test_config=None):
    """
    initialize Flask instance
    """
    _app = Flask(
        __name__,
        template_folder='templates',
    )

    # initialize config
    _app.config.from_object('{0}.settings'.format(__name__.split('.')[0]))
    _app.config.from_envvar('OATHENROLL_CFG', silent=False)

    _app.secret_key = _app.config['APP_SECRET']

    # initialize logging
    if _app.config['LOG_NAME'] is not None:
        _app.logger.name = _app.config['LOG_NAME']
    if _app.config['LOG_CONFIG'] is not None:
        if os.path.isfile(_app.config['LOG_CONFIG']):
            logging.config.fileConfig(_app.config['LOG_CONFIG'], disable_existing_loggers=True)
            _app.logger.removeHandler(flask.logging.default_handler)
            _app.logger.debug('Loaded logging config from %s', _app.config['LOG_CONFIG'])
        else:
            _app.logger.warning('No logging config file %s', _app.config['LOG_CONFIG'])
    if 'LOG_LEVEL' in os.environ:
        _app.logger.setLevel(os.environ['LOG_LEVEL'].upper())

    validate_config(_app.config)

    _app.template_folder = _app.config['TEMPLATES_DIRNAME']

    # URL routing
    _app.add_url_rule(
        '/',
        view_func=Default.as_view('default'),
        methods=('GET',),
    )
    _app.add_url_rule(
        '/reset',
        view_func=ResetToken.as_view('reset'),
        methods=('GET', 'POST'),
    )
    if _app.config['PROXY_LEVEL'] > 0:
        # see https://werkzeug.palletsprojects.com/en/1.0.x/middleware/proxy_fix/
        _app.logger.info('Using ProxyFix wrapper with %d proxy layers', _app.config['PROXY_LEVEL'])
        _app.wsgi_app = ProxyFix(
            _app.wsgi_app,
            x_for=_app.config['PROXY_LEVEL'],
            x_proto=_app.config['PROXY_LEVEL'],
            x_host=_app.config['PROXY_LEVEL'],
        )

    @_app.after_request
    def add_headers(response):
        """
        add HTTP headers to response
        """
        for header, value in HTTP_HEADERS:
            response.headers[header] = value
        return response

    return _app
